/*-
 *  Copyright (c) 2001-2003  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "unquote.h"
#include "uq_err.h"

static char cvs_id[] __UNUS =
"$Ringlet: c/misc/unquote/unquote.c,v 1.13 2003/07/07 08:55:06 roam Exp $";

char		uq_verstr[64] = "unquote";

/* max number of cmdline arguments allowed */
size_t		uq_arg_max = UQ_ARG_MAX;
/* index of max field (arg) to quote */
size_t		uq_argcnt;
/* a boolean array of flags for fields (args) to quote */
size_t		*uq_arg;

/* '-0': if '-n' was specified, separate arguments with a NUL character */
static int	zeroflag;
/* '-N': if '-n' was specified, print each argument on a separate line */
static int	Nflag;
/* '-n': do not actually execute anything, just print the expanded version */
static int	nflag;
/* '-p': do not search PATH, i.e. use execv() instead of execvp() */
static int	pflag;
/* '-R': do just the opposite: *quote* the listed arguments */
static int	Rflag;

static uq_err_t	uq_add_args(char *);
static uq_err_t	uq_add_arg(size_t);
static uq_err_t	uq_unquote_arg(char **, const char *);
static uq_err_t	uq_quote_arg(char **, const char *);
static uq_err_t	uq_fromhex(char *, const char *, size_t);
static uq_err_t	uq_tohex(char *, const char *, size_t);

static uq_err_t	uq_init(int, char *[]);
static uq_err_t	uq_doit(int, char *[]);

static void	usage(void);
static uq_err_t	version(void);

/*
 * Function:
 *         uq_makeversion()                - generate version string
 * Inputs:
 *         none
 * Returns:
 *         uq_err_t
 * Modifies:
 *         uq_verstr
 */

static uq_err_t
uq_makeversion(void) {

	snprintf(uq_verstr, sizeof(uq_verstr), "unquote v%d.%d"
#if UQ_VER_PRE
	    "-pre%d"
#endif /* UQ_VER_PRE */
#if UQ_VER_PATCH
	    "p%d"
#endif /* UQ_VER_PATCH */
	    ,
	    UQ_VER_MAJ, UQ_VER_MIN
#if UQ_VER_PRE
	    , UQ_VER_PRE
#endif /* UQ_VER_PRE */
#if UQ_VER_PATCH
	    , UQ_VER_PATCH
#endif /* UQ_VER_PATCH */
		);
	
	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_init				- general-purpose init function
 * Inputs:
 * 	argc, argv			- main() args for option processing
 * Returns:
 * 	uq_err_t
 * 	CMDLINE
 * Modifies:
 * 	verbose, quiet
 * 	may call uq_add_args()
 */

static uq_err_t
uq_init(int argc, char *argv[]) {
	int ch;		/* getopt() iterator over argv		*/
	int helpq;	/* boolean: '-h' on cmdline		*/
	int versq;	/* boolean: '-q' on cmdline		*/
	char *envargs;	/* UNQUOTE_ARGS env variable		*/
	uq_err_t r;	/* error code from called functions	*/
	
	uq_makeversion();
	
	/* Process environment variables */
	if (envargs = getenv("UNQUOTE_ARGS"), envargs != NULL)
		if (r = uq_add_args(envargs), r)
			return (r);

	/* Process cmdline options */
	
	helpq = versq = 0;
	opterr = 0;
	while (ch = getopt(argc, argv, UQ_OPTSTR), ch != EOF)
		switch (ch) {
			case '0':
				zeroflag = 1;
				break;
			case 'f':
				if (r = uq_add_args(optarg), r)
					return (r);
				break;
			case 'h':
				helpq = 1;
				break;
			case 'N':
				Nflag = 1;
				break;
			case 'n':
				nflag = 1;
				break;
			case 'p':
				pflag = 1;
				break;
			case 'q':
				quiet++;
				break;
			case 'R':
				Rflag = 1;
				break;
			case 'V':
				versq = 1;
				break;
			case 'v':
				verbose++;
				break;
			case '?':
			default:
				usage();
				/* NOTREACHED */
		}
	
	argc -= optind; argv += optind;
	
	if (versq)
		version();
	if (helpq)
		usage();
	if (versq || helpq)
		exit(UQ_ERR_NONE);
	
	/* sanity checks */
	if (argc < 1)
		usage();

	if (zeroflag && Nflag)
		usage();

	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	usage			- startup help info
 * Inputs:
 * 	none
 * Returns:
 * 	UQ_ERR_NONE
 * Modifies:
 * 	nothing, writes to stdout
 */

static void
usage(void) {
	static const char *msg[] = {
	    "Usage: unquote [-f fields] [-0 | -N] [-hnpqRVv] fields command "
	    "[args ...]",
	    "\t-0\twith -n, separate arguments with ASCII NUL characters;",
	    "\t-f\tlist arguments to expand;",
	    "\t-h\tprint this help text and exit;",
	    "\t-N\twith -n, separate arguments with newlines;",
	    "\t-n\tdo not execute command, just print out the expanded form;",
	    "\t-p\tdo not search PATH for the command;",
	    "\t-q\tquiet operation; multiple -q's make it even more quiet;",
	    "\t-R\treverse operation: quote the specified arguments;"
	    "\t-V\tprint version information and exit;",
	    "\t-v\tverbose operation; multiple -v's increase verbosity level.",
	    NULL
	};
	unsigned i;

	for (i = 0; msg[i] != NULL; i++)
		fprintf(stderr, "%s\n", msg[i]);
	exit(UQ_ERR_CMDLINE);
}

/*
 * function:
 * 	version			- output version info
 * Inputs:
 * 	none
 * Returns:
 * 	UQ_ERR_NONE
 * Modifies:
 * 	nothing, writes to stdout
 */

static uq_err_t
version(void) {
	
	printf("%s\n", uq_verstr);
	
	if (verbose) {
		printf("%s\n", "Built on " __DATE__ ", " __TIME__);
#ifdef __GNUC__
		if (verbose > 1)
			printf("%s\n", "Compiler: GNU C " __VERSION__
			    " on " UQ_OS " " UQ_OSREL ": " UQ_OSHOST);
#endif /* __GNUC__ */
	}

	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_doit				- perform the actual work
 * Inputs:
 * 	argc, argv			- main() args
 * Returns:
 * 	uq_err_t
 * 	NOMEM				- allocationg the new cmdline args
 * 	EXEC				- the final exec*() fails
 * 	error code from uq_add_args() or uq_unquote_arg()
 * Modifies:
 * 	calls uq_add_args() and uq_unquote_arg()
 * 	calls execv() or execvp() at the end and never returns on success
 */

static uq_err_t
uq_doit(int argc, char *argv[]) {
	size_t i;		/* iterator over argv/argc		*/
	char **nargv;		/* argv array with expanded arguments	*/
	char sepchar;		/* '-n' args separator character	*/
	uq_err_t (*quote_fun)(char **, const char *);
	uq_err_t r;		/* error code from called functions	*/
	
	/* Get additional fields to quote from the first cmdline arg */
	if (r = uq_add_args(argv[0]), r)
		return (r);
	argc--;
	argv++;

	if (!Rflag)
		quote_fun = uq_unquote_arg;
	else
		quote_fun = uq_quote_arg;

	/* Store and quote as needed */
	if (nargv = calloc(argc+1, sizeof(*nargv)), nargv == NULL)
		return (UQ_ERR_NOMEM);
	for (i = 0; i < (size_t) argc; i++)
		if ((i < uq_argcnt) && uq_arg[i]) {
			if (r = quote_fun(nargv+i, argv[i]), r)
				return (r);
		} else {
			nargv[i] = strdup(argv[i]);
		}

	if (nflag) {
		sepchar = zeroflag? '\0': Nflag? '\n': ' ';
		for (i = 0; i < (size_t) argc; i++) {
			printf("%s", nargv[i]);
			if (i < (size_t) argc - 1)
				putchar(sepchar);
			else if (!zeroflag)
				putchar('\n');
		}
		return (UQ_ERR_NONE);
	}
	
	if (!pflag)
		execvp(nargv[0], nargv);
	else
		execv(nargv[0], nargv);
	return (UQ_ERR_EXEC);
}

/*
 * Function:
 * 	uq_add_args			- parse a list of fields to quote
 * Inputs:
 * 	s				- field list
 * Returns:
 * 	uq_err_t
 * 	UQ_ERR_CMDLINE			- bad format for the field list
 * 	error code from uq_add_arg()
 * Modifies:
 * 	nothing by itself; calls uq_add_arg() for each field in the list
 */

static uq_err_t
uq_add_args(char *s) {
	char *p;	/* iterator over s			*/
	long v;		/* numeric value parsed from s		*/
	size_t r;	/* error code from called functions	*/

	if (s == NULL)
		return (UQ_ERR_NONE);

	while (*s) {
		v = strtol(s, &p, 10);
		if ((s == p) || (v < 0))
			return (UQ_ERR_CMDLINE);
		if (r = uq_add_arg((size_t) v), r)
			return (r);
		s = p;
		if (*s != '\0') {
			if (*s == ':')
				s++;
			else
				return (UQ_ERR_CMDLINE);
		}
	}

	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_add_arg()			- store a single field to quote
 * Inputs:
 * 	v				- field index
 * Returns:
 * 	uq_err_t
 * 	CMDLINE				- index too big
 * 	NOMEM				- reallocatiing uq_args[]
 * Modifies:
 * 	uq_args[]
 * 	may reallocate (grow) uq_args[] and update uq_argcnt
 */

static uq_err_t
uq_add_arg(size_t v) {

	if (v > uq_arg_max)
		return (UQ_ERR_CMDLINE);

	if (v >= uq_argcnt) {
		size_t newcnt, *narg;

		newcnt = (size_t) v+1;
		narg = realloc(uq_arg, newcnt * sizeof(*uq_arg));
		if (narg == NULL)
			return (UQ_ERR_NOMEM);
		memset(narg + uq_argcnt, 0,
		    (newcnt - uq_argcnt) * sizeof(*narg));
		uq_arg = narg;
		uq_argcnt = newcnt;
	}

	uq_arg[v] = 1;
	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_fromhex			- expand a hex value into a string
 * Inputs:
 * 	dst				- val/res destination string
 * 	src				- source (hex) string
 * 	len				- length of hex value
 * Returns:
 * 	uq_err_t
 * 	BADCHAR				- a non-hex-digit char in src
 * Modifies:
 * 	dst[]
 */

static uq_err_t
uq_fromhex(char *dst, const char *src, size_t len) {
	size_t val;	/* the numeric value of src	*/
	size_t i;	/* iterator over src		*/

	val = 0;
	for(i = 0; i < len; i++) {
		if (!isxdigit(src[i]))
			return (UQ_ERR_BADCHAR);
		if (isdigit(src[i]))
			val = (val * 16) + src[i] - '0';
		else
			val = (val * 16) + toupper(src[i]) - 'A' + 10;
	}

	*dst = val;
	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_tohex			- encode a string into a hex value
 * Inputs:
 * 	dst				- val/res destination hex value
 * 	src				- source string
 * 	len				- source string length
 * Returns:
 * 	uq_err_t
 * 	nothing so far
 * Modifies:
 * 	dst[]
 */

static uq_err_t
uq_tohex(char *dst, const char *src, size_t len) {
	size_t i;
	
	for (i = 0; i < len; i++)
		sprintf(dst+2*i, "%02X", src[i]);
	/* sprintf() null-terminates the result for us. */
	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_unquote_arg			- unquote (expand) a field
 * Inputs:
 * 	pdst				- val/res pointer to expanded string
 * 	src				- string to unquote
 * Returns:
 * 	uq_err_t
 * 	NOMEM				- allocating expanded string
 * 	error code from uq_fromhex()
 * Modifies:
 * 	allocates and stores into *pdst
 */

static uq_err_t
uq_unquote_arg(char **pdst, const char *src) {
	char *dst;	/* the real unquoted string		*/
	const char *p;	/* iterator over src			*/
	char *q;	/* iterator over dst			*/
	size_t l;	/* length of src			*/
	uq_err_t r;	/* error code from called functions	*/

	l = strlen(src);
	/* the (l%2) is there so that '3bf' is equiv to '03bf' */
	if (dst = calloc(l + (l%2) + 1, 1), dst == NULL)
		return (UQ_ERR_NOMEM);
	p = src;
	q = dst;
	if (l%2) {
		if (r = uq_fromhex(q++, p++, 1), r) {
			free(dst);
			return (r);
		}
	}
	while (*p) {
		if (r = uq_fromhex(q, p, 2), r) {
			free(dst);
			return (r);
		}
		q += 1;
		p += 2;
	}

	*pdst = dst;
	return (UQ_ERR_NONE);
}

/*
 * Function:
 * 	uq_quote_arg			- quote (encode in hex) a field
 * Inputs:
 * 	pdst				- val/res pointer to hex string
 * 	src				- string to quote
 * Returns:
 * 	uq_err_t
 * 	NOMEM				- allocating hex string
 * 	error code from uq_tohex()
 * Modifies:
 * 	allocates and stores into *pdst
 */

static uq_err_t
uq_quote_arg(char **pdst, const char *src) {
	char *dst;	/* the real quoted string		*/
	size_t l;	/* length of src			*/
	uq_err_t r;	/* error code from called functions	*/

	l = strlen(src);
	if (dst = calloc(2*l + 1, 1), dst == NULL)
		return (UQ_ERR_NOMEM);
	if (r = uq_tohex(dst, src, l), r) {
		free(dst);
		return (r);
	}

	*pdst = dst;
	return (UQ_ERR_NONE);
}

/*
 *   M A I N   F U N C T I O N
 */

int
uq_main(int argc, char *argv[]) {
	uq_err_t r;	/* error code from called functions	*/
	
	if (r = uq_init(argc, argv), r)
		return (uq_prerror("init", r));
	
	argc -= optind;
	argv += optind;
	if (r = uq_doit(argc, argv), r)
		uq_prerror("doit", r);
	
	return (r);
}
